/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents the common traits of BPEL activities.
 * 
 * @author Greg Hackmann
 */
abstract class Activity
{
	/**
	 * The parser that the input is coming from.
	 */
	protected final XmlPullParser parser;
	/**
	 * The actual name of the tag (e.g., "assign", "copy", etc.)
	 */
	protected final String type;
	
	/**
	 * The tag's name attribute.
	 */
	protected String name;	
	/**
	 * The tag's suppressJoinFailure attribute.
	 */
	protected boolean suppressJoinFailure = false;
	/**
	 * The tag's joinCondition attribute.
	 */
	protected String joinCondition = null;
	
	/**
	 * The scope enclosing the tag.
	 */
	protected final ScopeData scopeData;
	/**
	 * The tag's target links.
	 */
	protected Target [] targets;
	/**
	 * The tag's source links.
	 */
	protected Source [] sources;

	/**
	 * Creates a new Activity.
	 * 
	 * @param parser the parser to read the tag's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * @param tagName the type of the activity tag
	 */
	protected Activity(XmlPullParser parser, ScopeData scopeData,
		String tagName)
	{
		this.type = tagName;
		this.parser = parser;
		this.scopeData = scopeData;
	}

	/**
	 * Parses the next activity tag.  
	 * 
	 * @param parser the parser to read the tag's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * @return the next activity tag that the parser read
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	static Activity parse(XmlPullParser parser, ScopeData scopeData) throws
		IOException, XmlPullParserException, MalformedDocumentException
	{
		String type = parser.getName();
		// Get the name of the next tag
		
		if("receive".equals(type))
			return new Receive(parser, scopeData);
		else if("reply".equals(type))
			return new Reply(parser, scopeData);
		else if("sequence".equals(type))
			return new Sequence(parser, scopeData);
		else if("flow".equals(type))
			return new Flow(parser, scopeData);
		else if("assign".equals(type))
			return new Assign(parser, scopeData);
		else if("invoke".equals(type))
			return new Invoke(parser, scopeData);
		else if("switch".equals(type))
			return new Switch(parser, scopeData);
		else if("while".equals(type))
			return new While(parser, scopeData);
		else if("empty".equals(type))
			return new Empty(parser, scopeData);
		else if("terminate".equals(type))
			return new Terminate(parser, scopeData);
		else if("scope".equals(type))
			return new Scope(parser, scopeData);
		else if("throw".equals(type))
			return new Throw(parser, scopeData);
		else if("wait".equals(type))
			return new Wait(parser, scopeData);
		else if("pick".equals(type))
			return new Pick(parser, scopeData);
		// Delegate the parsing to the appropriate class based on the tag name
		
		// TODO: compensate
		
		throw new MalformedBPELException(parser,
			"<" + type + "> is not a BPEL activity");
	}
	
	/**
	 * Fires the tag's source links.
	 * 
	 * @param instanceData information about the current process instance
	 * @param succeeded whether or not the tag's execution was successful
	 * 
	 * @throws BPELServerFault one of the link's status could not be
	 * evaluated
	 */
	protected void fireLinks(InstanceData instanceData, boolean succeeded)
		throws BPELServerFault
	{
		int i = 0;
		try
		{
			for(i = 0; i < sources.length; i++)
				sources[i].fire(instanceData, succeeded);
			// Try to fire all the links
		}
		catch(BPELServerFault e)
		{
			for(int j = i + 1; j < sources.length; j++)
				sources[j].fire(instanceData, false);
			
			throw e;
			// If any one fails, then fire the remaining links as false, and
			// re-throw the exception we caught.
			// (Note that this set of links won't throw an exception while
			// firing, since the transitionCondition is never evaulated
			// when success == false.)
		}
	}
	
	/**
	 * Waits on incoming links, executes the tag, and fires outgoing links.
	 * Note that most of the work is done by
	 * each class's implementation of {@link #executeImpl(InstanceData)};
	 * this method just does the bookkeeping regarding the links.
	 * 
	 * @param instanceData information about the current process instance
	 * @throws BPELServerFault an uncaught fault was thrown during
	 * the tag's execution
	 * 
	 * @return whether or not the tag executed successfully
	 */
	boolean execute(InstanceData instanceData) throws BPELServerFault
	{
		Hashtable linkStatus = new Hashtable();
		// For each of the target links
		for(int i = 0; i < targets.length; i++)
		{
			LinkSpecification linkSpec = targets[i].getLinkSpecification();
			ConcreteLink link = instanceData.getConcreteLink(linkSpec);
			// Get the link's instantiation

			BPELServer.log.info(toString() + ": " + "Waiting for link " +
				linkSpec + " to fire");
			link.waitForStatus();
			BPELServer.log.info(toString() + ": " + "Link " + linkSpec +
				" fired");
			linkStatus.put(linkSpec.getName(), new Boolean(link.getStatus()));
			// Wait for the link to fire and store its status
		}
		
		Enumeration e = linkStatus.elements();
		boolean shouldExecute = targets.length == 0;
		while(e.hasMoreElements())
		{
			Boolean condition = (Boolean)e.nextElement();
			shouldExecute |= condition.booleanValue();
		}
		// TODO: use joinCondition
		
		// If the joinCondition failed
		if(!shouldExecute)
		{
			String error = "joinCondition failed at " + this;
			// If we're not supposed to suppress the failure
			if(!suppressJoinFailure)
			{
				BPELServer.log.severe(toString() + ": " + error +
					"; fault thrown");

				// Check to see if we can handle a joinFailure fault
				if(!instanceData.handleFault(scopeData, BPELServer.namespace,
					"joinFailure"))
				{
					Exception ex = new Exception(error);
					fireLinks(instanceData, false);
					throw new BPELServerFault(ex);
					// If not, fire all the links as false and propagate the
					// fault all the way up to the top
				}
				
				fireLinks(instanceData, false);
				return false;
				// Otherwise, just quietly fire the links and stop
			}
			
			BPELServer.log.info(toString() + ": " + error +
				"; fault suppressed");
			fireLinks(instanceData, false);
			return false;
			// If we're supposed to suppress it, log it as such, then quietly
			// fail
		}
		
		boolean succeeded = false;
		try
		{
			succeeded = executeImpl(instanceData);
		}
		finally
		{
			fireLinks(instanceData, succeeded);
		}
		// Try to actually execute the tag, and fire the links based on its
		// success
		
		return succeeded;
	}
	
	/**
	 * Executes the tag.
	 * 
	 * @param instanceData information about the current process instance
	 * @throws BPELServerFault an uncaught fault was thrown during
	 * the tag's execution
	 * 
	 * @return whether or not the tag executed successfully
	 * 
	 * Note: the only times that this method should return <tt>false<tt> rather
	 * than throwing a fault is when &lt;terminate&gt; is executed but caught,
	 * or when &lt;sequence&gt; has a child &lt;terminate&gt; like that.
	 * This means that the process should not come to a griding halt, but
	 * dead path elimination should take place.
	 */
	abstract boolean executeImpl(InstanceData instanceData) throws
		BPELServerFault;
	
	/**
	 * Gets a list of any "start" activities (i.e., receive activities with
	 * createInstance="true") contained within this tag.  Generally, most
	 * tags will return an empty list; those that do not (e.g., {@link Receive})
	 * should override this method.
	 * 
	 * @return the list of start activities
	 */
	Transaction [] getStartActivities()
	{
		return new Transaction[0];
	}
	
	/**
	 * Gets the value of an attribute from the parser
	 * 
	 * @param attribute the name of the attribute
	 * @return the value of the attribute, or <tt>null</tt> if the tag has
	 * no such attribute
	 */
	protected final String getAttribute(String attribute)
	{
		return parser.getAttributeValue(null, attribute);
	}
	
	/**
	 * Parses the opening tag, including all standard attributes.
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedBPELException the BPEL parser read malformed data
	 */
	protected void parseStartTag() throws XmlPullParserException,
		IOException, MalformedBPELException
	{
		parser.require(XmlPullParser.START_TAG, BPELServer.namespace, type);
		
		name = getAttribute("name");
		joinCondition = getAttribute("joinCondition");
		suppressJoinFailure = scopeData.isSuppressJoinFailure() ||
			"true".equals(getAttribute("suppressJoinFailure"));
	}
	
	/**
	 * Parses standard child tags.
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	protected void parseStandardElements() throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		Vector targetVector = new Vector();
		while("target".equals(parser.getName()))
			targetVector.addElement(new Target(parser, this));
		
		targets = new Target[targetVector.size()];
		targetVector.copyInto(targets);

		Vector sourceVector = new Vector();
		while("source".equals(parser.getName()))
			sourceVector.addElement(new Source(parser, scopeData, this));

		sources = new Source[sourceVector.size()];
		sourceVector.copyInto(sources);
	}
	

	/**
	 * Parses the closing tag.
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 */
	protected void parseEndTag() throws XmlPullParserException, IOException
	{
		parser.require(XmlPullParser.END_TAG, BPELServer.namespace, type);
	}

	/**
	 * Gets the activity's name, if any.
	 * 
	 * @return the activity's name, or <tt>null</tt> if the activity is unnamed
	 */
	public String getName()
	{
		return name;
	}
	
	public String toString()
	{
		if(name == null)
			return "<" + type + ">";
		return name;
	}
}
