/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.kxml2.io.*;
import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.soap.*;
import edu.wustl.mobilab.sliver.util.*;

/**
 * A server that handles SOAP requests.
 * 
 * <p>Example usage:</p>
 * 
 * <code>
 * String namespace = "http://some/namespace";
 * <br />SocketTransport transport = new SocketTransport(9000);
 * <br />BPELServer server = new BPELServer(transport);
 * <br />server.addProcess(namespace, new FileInputStream("SomeWorkflow.bpel")); 
 * <br />server.bindIncomingLink("ClientLink", namespace, "ExecuteWorkflow");
 * <br />server.bindOutgoingLink("ExternalSOAPServiceLink", new SocketAddress("some.host", 9000));
 * <br />server.start();
 * </code>
 * 
 * @see SOAPServer
 * 
 * @author Greg Hackmann
 */
public class BPELServer extends SOAPServer
{
	/**
	 * The BPEL namespace.
	 */
	public static final String namespace =
		"http://schemas.xmlsoap.org/ws/2003/03/business-process/";
	
	/**
	 * The {@link Logger} that will log BPEL server events.
	 */
	public static final Logger log =
		Logging.getLogger("edu.wustl.mobilab.sliver.util.bpel");

	/**
	 * All known BPEL processes.
	 */
	private final Hashtable processes = new Hashtable();
	/**
	 * The bindings for incoming partner links.
	 */
	private final Hashtable incomingBindings = new Hashtable();
	/**
	 * The bindings for outgoing partner links.
	 */
	private final Hashtable outgoingBindings = new Hashtable();


	/**
	 * Creates a new BPEL server.
	 * 
	 * @param transport the transport that SOAP requests will arrive from
	 */
	public BPELServer(Transport transport)
	{
		super(transport);
	}

	/**
	 * Binds an incoming partner link to an operation namespace/name.
	 * 
	 * @param partnerLinkName the partner link's name
	 * @param operationNamespace the namespace of the partner link's SOAP operation
	 * @param operation the partner link's SOAP operation
	 */
	public void bindIncomingLink(String partnerLinkName, String operationNamespace, String operation)
	{
		incomingBindings.put(operationNamespace + ":" + operation, partnerLinkName);
	}
	
	/**
	 * Gets an incoming partner link binding.
	 * 
	 * @param operationNamespace the namespace of the partner link's SOAP operation
	 * @param operation the partner link's SOAP operation
	 *
	 * @return the name of the corresponding partner link
	 */
	String getIncomingLink(String operationNamespace, String operation)
	{
		return (String)incomingBindings.get(operationNamespace + ":" + operation);
	}

	/**
	 * Binds an outgoing partner link.
	 * 
	 * @param partnerLinkName the partner link's name
	 * @param binding the partner link's binding
	 */
	public void bindOutgoingLink(String partnerLinkName, Object binding)
	{
		outgoingBindings.put(partnerLinkName, binding);
	}

	/**
	 * Gets an outgoing partner link binding.
	 * 
	 * @param partnerLinkName the partner link's name
	 * @return the corresponding binding
	 */
	Object getOutgoingBinding(String partnerLinkName)
	{
		return outgoingBindings.get(partnerLinkName);
	}

	/**
	 * Adds a process to the BPEL server.
	 * 
	 * @param operationNamespace the process's namespace
	 * @param in an {@link InputStream} containing the process's BPEL
	 * specification
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	public void addProcess(String operationNamespace, InputStream in)
		throws XmlPullParserException, IOException, MalformedDocumentException
	{
		XmlPullParser parser = new KXmlParser();
		parser.setInput(in, null);
		parser.setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES, true);
		// Create the XML parser
		
		parser.require(XmlPullParser.START_DOCUMENT, null, null);
		parser.nextTag();
		// Parse the start of the document

		Process process = new Process(parser);
		// Parse the process
		
		parser.next();
		parser.require(XmlPullParser.END_DOCUMENT, null, null);	
		// Parse the end of the document
		
		processes.put(operationNamespace, process);
		// Store our new process for later execution
	}
	
	protected Object receiveRequest(Channel channel) throws IOException,
		MustUnderstandException, XmlPullParserException
	{
		return channel.receiveElement();
	}
	
	protected void handleRequest(Object request, Channel channel)
		throws IOException
	{
		Element soapRequest = (Element)request;
		
		String requestNamespace = soapRequest.getNamespace();
		Process process = (Process)processes.get(requestNamespace);
		// Get the request's namespace and look up the process
		
		if(process == null)
		{
			Exception e = new IllegalArgumentException("Unknown namespace "
				+ requestNamespace);
			sendFault("Client", e.getMessage(), e.getMessage(), channel);
			return;
		}
		// Complain if the process doesn't exist
		
		try
		{
			process.execute(this, soapRequest, channel);
		}
		// Try to execute the process
		catch(BPELServerFault e)
		{
			log.severe("Sending uncaught BPELServerFault " + e);
			channel.sendObject(e.getFault());

			if(!channel.isClosed())
			{
				log.severe("Closing channel " + channel);
				channel.close();
			}
		}
		// Send back any BPEL faults that occur while executing the process
		catch(Throwable t)
		{
			log.severe("Exception thrown while executing process: " + t);
			t.printStackTrace();
			
			BPELServerFault fault = new BPELServerFault(t);
			log.severe("Sending uncaught Exception " + fault);
			channel.sendObject(fault.getFault());
			
			if(!channel.isClosed())
			{
				log.severe("Closing channel " + channel);
				channel.close();
			}
		}
		// Wrap any Java faults and send those back.  This should generally
		// not happen, since Sliver should wrap Java faults in BPEL faults
		// after they're thrown, but we'll catch them anyway just in case.
	}
	
	public String toString()
	{
		return "BPELServer: " + transport.toString();
	}
}
