/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;flow&gt; tags.
 * 
 * @author Greg Hackmann
 */
class Flow extends Activity
{
	/**
	 * The new links declared within the &lt;flow&gt;'s scope.
	 */
	private final Links links;
	/**
	 * A list of all the child activities.
	 */
	private final Vector children = new Vector();
	
	/**
	 * Creates a new Flow.
	 * 
	 * @param parser the parser to read the flow's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Flow(XmlPullParser parser, ScopeData scopeData) throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		super(parser, scopeData, "flow");
		parseStartTag();
		
		parser.nextTag();
		parseStandardElements();
		
		links = new Links(parser);
		// Parse the new links
		
		ScopeData newScopeData;
		if(links.isEmpty())
			newScopeData = scopeData;
		// If there aren't any, then reuse the old scope
		else
		{
			newScopeData = new ScopeData(scopeData);
			newScopeData.addLinks(links);
		}
		// Otherwise, copy the old scope, and add the new links to the copy
		
		while(parser.getEventType() != XmlPullParser.END_TAG)
			children.addElement(Activity.parse(parser, newScopeData));
		// Parse the children activities
		
		if(children.isEmpty())
			throw new MalformedBPELException(parser,
				"<flow> must specify at least one activity");
		// Complain if there are no children
		
		parseEndTag();
		parser.nextTag();
	}

	boolean execute(InstanceData instanceData) throws BPELServerFault
	{
		links.createConcreteLinks(instanceData);
		return super.execute(instanceData);
		// Set up the new links before executing the process
	}
	
	boolean executeImpl(InstanceData instanceData) throws BPELServerFault
	{		
		FlowThread [] childThreads = new FlowThread[children.size()];
		// For each child activity
		for(int i = 0; i < childThreads.length; i++)
		{
			Activity activity = (Activity)children.elementAt(i);
			String threadName = this + " child " + activity + " (" + i + ")";
			
			childThreads[i] = new FlowThread(instanceData, activity, threadName);
			childThreads[i].start();
			// Create a new thread containing that child
		}
		
		// For each thread
		for(int i = 0; i < childThreads.length; i++)
		{
			try
			{
				childThreads[i].waitForCompletion();
			}
			catch(InterruptedException e)
			{
				throw new BPELServerFault(e);
			}
			// Wait for the thread to complete
		}
		
		return true;
	}
	
	Transaction [] getStartActivities()
	{
		Vector startActivities = new Vector();
		
		Enumeration e = children.elements();
		// For each child activity
		while(e.hasMoreElements())
		{
			Activity next = (Activity)e.nextElement();
			Transaction [] nextStart = next.getStartActivities();
			for(int i = 0; i < nextStart.length; i++)
				startActivities.addElement(nextStart[i]);
			// Add its start activities
		}
		
		Transaction [] startActivitiesArray = new Transaction[startActivities.size()];
		startActivities.copyInto(startActivitiesArray);
		return startActivitiesArray;
		// Return the start activities in array form
	}
}

/**
 * Executes one of a {@link Flow}'s child activities.
 * 
 * @author Greg Hackmann
 */
class FlowThread extends Thread
{
	/**
	 * The current BPEL process instance.
	 */
	private final InstanceData instanceData;
	/**
	 * The activity that should be executed
	 */
	private final Activity child;
	/**
	 * The fault that the child threw, if any.
	 */
	private BPELServerFault fault;
	
	/**
	 * Creates a new FlowThread.
	 * 
	 * @param instanceData the current BPEL process instance
	 * @param child the activity that should be executed
	 * @param name the thread's name
	 */
	FlowThread(InstanceData instanceData, Activity child, String name)
	{
		super(name);
		this.instanceData = instanceData;
		this.child = child;
	}
	
	public void run()
	{
		try
		{
			child.execute(instanceData);
		}
		catch(BPELServerFault e)
		{
			fault = e;
		}
		// Execute the child, and save any faults for later
	}
	
	/**
	 * Waits for the child activity to complete.
	 * 
	 * @throws BPELServerFault there was an error while executing the
	 * activity
	 * @throws InterruptedException the activity was interrupted before
	 * it could complete
	 */
	void waitForCompletion() throws BPELServerFault, InterruptedException
	{
		join();
		if(fault != null)
			throw fault;
		// Wait for the thread to die, and re-throw any faults
	}
}
