/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;from&gt; tags.
 * 
 * @author Greg Hackmann
 */
class From
{
	/**
	 * The source of the data that will be read.
	 */
	private final FromSource source;

	/**
	 * Creates a new From.
	 * 
	 * @param parser the parser to read the from's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */	
	From(XmlPullParser parser, ScopeData scopeData)
		throws XmlPullParserException, IOException, MalformedDocumentException
	{
		parser.require(XmlPullParser.START_TAG, BPELServer.namespace, "from");
		
		if(hasAttribute(parser, "variable"))
			source = new VariablePartSource(parser, scopeData);
		// First try variable sources
		else if(hasAttribute(parser, "expression"))
		{
			String expression = parser.getAttributeValue(null, "expression");
			source = new ExpressionSource(expression, parser, scopeData);
		}
		// Then try XPath expression sources
		else
		{
			String text = parser.nextText();
			if(text.trim().length() > 0)
				source = new LiteralSource(text);
			else
				throw new MalformedBPELException(parser,
					"<from> has invalid from-spec");
		}
		// Finally assume the text represents a literal source
		
		// TODO:
		// <from partnerLink="ncname" endpointReference="myRole|partnerRole"/>
		// <from variable="ncname" property="qname"/>
		
		parser.require(XmlPullParser.END_TAG, BPELServer.namespace, "from");
		parser.nextTag();
	}
	
	/**
	 * Reads the data from the specified source.
	 * 
	 * @param instanceData the current BPEL process instance
	 * @return an {@link Element} representation of the data that was read
	 * @throws BPELServerFault an error occured while reading the data
	 */
	Element read(InstanceData instanceData) throws BPELServerFault
	{
		return source.read(instanceData);
	}
	
	/**
	 * Determines whether or not the tag has the specified attribute.
	 * 
	 * @param parser the parser that is reading the current tag
	 * @param name the name of the attribute to look for
	 * @return whether or not the tag has the specified attribute
	 */
	private boolean hasAttribute(XmlPullParser parser, String name)
	{
		return parser.getAttributeValue(null, name) != null;
	}
}