/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;
import edu.wustl.mobilab.sliver.xpath.*;

/**
 * Describes the source of data for a {@link From} operation.
 * 
 * @author Greg Hackmann
 */
interface FromSource
{
	/**
	 * Reads data from the current source.
	 * 
	 * @param instanceData the current BPEL process instance
	 * @return an {@link Element} representation of the data that was read
	 * 
	 * @throws BPELServerFault an error occurred while reading the data
	 */
	Element read(InstanceData instanceData) throws BPELServerFault;
}

/**
 * Describes data sources specified by an XPath expression.
 * 
 * @author Greg Hackmann
 */
class ExpressionSource implements FromSource
{
	/**
	 * The expression that specifies the data source.
	 */
	private final Expression expression;
	/**
	 * The process's current scope.
	 */
	private final ScopeData scopeData;
	
	/**
	 * Creates a new ExpressionSource.
	 * 
	 * @param expression a String-encoded XPath expression
	 * @param parser the parser that the expression came from
	 * @param scopeData the process's current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedExpressionException the XPath parser read malformed data
	 */
	ExpressionSource(String expression, XmlPullParser parser, ScopeData scopeData)
		throws MalformedExpressionException, IOException, XmlPullParserException
	{
		this.expression = ExpressionParser.parse(expression, parser, scopeData);
		this.scopeData = scopeData;
		// Parse the expression and save the scope for later
		
		parser.nextTag();
	}
	
	public Element read(InstanceData instanceData) throws BPELServerFault
	{
		try
		{
			Object result = expression.evaluate(scopeData, instanceData);
			// Evaluate the XPath expression

			Element element = new Element();
			element.addChild(Node.TEXT, result.toString());
			return element;
			// Save the result in an Element
			// FIXME: I can't imagine that this will work for complex data
			// types.
		}
		catch(ExpressionEvaluationException e)
		{
			throw new BPELServerFault(e);
		}
	}
}

/**
 * Describes data sources specified by literals.
 * 
 * @author Greg Hackmann
 */
class LiteralSource implements FromSource
{
	/**
	 * The literal data source.
	 */
	private final String literal;

	/**
	 * Creates a new LiteralSource.
	 * 
	 * @param literal the literal data source
	 */
	LiteralSource(String literal)
	{
		this.literal = literal;
	}
	
	public Element read(InstanceData instanceData) throws BPELServerFault
	{
		Element element = new Element();
		element.addChild(Node.TEXT, literal);
		// Create a new Element with the literal as its text
		
		return element;
	}	
}

/**
 * Describes data sources specified by BPEL variable parts.
 * 
 * @author Greg Hackmann
 */
class VariablePartSource implements FromSource
{
	/**
	 * The BPEL variable that the data will be read from.
	 */
	private final VariableSpecification variable;
	/**
	 * The part of the variable that will be read.
	 */
	private final String partName;
	/**
	 * The process's current scope.
	 */
	private final ScopeData scopeData;
	
	/**
	 * Creates a new VariablePartSource.
	 * 
	 * @param parser the parser that the expression comes from
	 * @param scopeData the process's current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedBPELException the BPEL parser read malformed data
	 */
	VariablePartSource(XmlPullParser parser, ScopeData scopeData)
		throws MalformedBPELException, IOException, XmlPullParserException
	{
		this.scopeData = scopeData;
		
		String part = parser.getAttributeValue(null, "part");
		if(part == null)
			partName = null;
		else
		{
			String [] splitPart = Namespace.expandNamespace(part, parser);
			partName = splitPart[1];
		}
		// Read the part, and split off the name portion if the part isn't empty
		
		String variableName = parser.getAttributeValue(null, "variable");
		if(variableName == null)
			throw new MalformedBPELException(parser,
				"<from> must specify variable");
		variable = scopeData.getVariable(variableName);
		// Read the variable name and get the corresponding spec		
		
		parser.nextTag();
	}
	
	public Element read(InstanceData instanceData) throws BPELServerFault
	{
		return instanceData.getVariablePart(scopeData, variable, partName, null);
	}	
}
