/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.util.*;

import org.kxml2.kdom.*;

import edu.wustl.mobilab.sliver.soap.*;

/**
 * Represents a BPEL process instance.
 * 
 * @author Greg Hackmann
 */
public class InstanceData
{
	/**
	 * The values of variables.
	 */
	private Hashtable variables = new Hashtable();
	/**
	 * The status of inter-activity links.
	 */
	private Hashtable concreteLinks = new Hashtable();
	/**
	 * The concrete bindings for each partner link.
	 */
	private Hashtable partnerLinkBindings = new Hashtable();
	/**
	 * A list of all enabled partner links.
	 */
	private Vector enabledPartnerLinks = new Vector();
	
	/**
	 * The {@link BPELServer} that is hosting the process.
	 */
	private final BPELServer server;
	
	/**
	 * Creates a new InstanceData.
	 * 
	 * @param server the {@link BPELServer} that is hosting the process
	 */
	InstanceData(BPELServer server)
	{
		this.server = server;
	}
	
	/**
	 * Gets the {@link BPELServer}'s transport.
	 * 
	 * @return the {@link BPELServer}'s transport
	 */
	Transport getTransport()
	{
		return server.getTransport();
	}
	
	/**
	 * Gets the value of a variable.
	 * 
	 * @param variableSpec the variable's specification
	 * @return an {@link Element} representing the variable's value, or
	 * <tt>null</tt> if it is unset
	 */
	public Element getVariable(VariableSpecification variableSpec)
	{
		return getVariable(variableSpec.getName());
	}
	/**
	 * Gets the value of a variable.
	 * 
	 * @param name the variable's name
	 * @return an {@link Element} representing the variable's value, or
	 * <tt>null</tt> if it is unset
	 */	
	Element getVariable(String name)
	{
		return (Element)variables.get(name);
	}

	/**
	 * Sets the value of a variable.
	 * 
	 * @param variableSpec the variable's specification
	 * @param value an {@link Element} representing the variable's new value
	 */
	void setVariable(VariableSpecification variableSpec, Element value)
	{
		setVariable(variableSpec.getName(), value);
	}
	
	/**
	 * Sets the value of a variable.
	 * 
	 * @param name the variable's name
	 * @param value an {@link Element} representing the variable's new value
	 */
	void setVariable(String name, Element value)
	{
		variables.put(name, value);
	}
	
	/**
	 * Gets the instantiation of an inter-activity link.
	 * 
	 * @param linkSpec the link's specification
	 * @return the corresponding inter-activity link
	 */
	ConcreteLink getConcreteLink(LinkSpecification linkSpec)
	{
		return (ConcreteLink)concreteLinks.get(linkSpec.getName());
	}
	
	/**
	 * Sets the instantiation of an inter-activity link.
	 * 
	 * @param linkSpec the link's specification
	 * @param link the instantiated inter-activity link
	 */
	void setConcreteLink(LinkSpecification linkSpec, ConcreteLink link)
	{
		concreteLinks.put(linkSpec.getName(), link);
	}
	
	/**
	 * Waits for any one of a set of {@link OnMessage}s to receive a message.
	 * 
	 * @param onMessages the {@link OnMessage}s to listen for
	 * @return the {@link OnMessage} which received its message first
	 * 
	 * @throws InterruptedException the thread was interrupted while waiting
	 * for a message to arrive
	 */
	synchronized OnMessage waitForMessage(OnMessage [] onMessages)
		throws InterruptedException
	{
		while(true)
		{
			// For each onMessage
			for(int i = 0; i < onMessages.length; i++)
			{
				PartnerLink link = onMessages[i].partnerLink;
				// Get the onMessage's partner link
				if(partnerLinkBindings.containsKey(link.getName()))
					return onMessages[i];
				// If it's bound (i.e., an incoming message has been sent to
				// it), then we're done
			}

			wait();
			// Sleep until a new partner link is bound
		}		
	}

	/**
	 * Binds a partner link.
	 * 
	 * @param partnerLink the partner link to bind
	 * @param binding the {@link Channel} to bind the link to
	 */
	synchronized void bindPartnerLink(PartnerLink partnerLink, Channel binding)
	{
		bindPartnerLink(partnerLink.getName(), binding);
	}
		
	/**
	 * Binds a partner link.
	 * 
	 * @param partnerLinkName the name of the partner link to bind
	 * @param binding the {@link Channel} to bind the link to
	 */
	synchronized void bindPartnerLink(String partnerLinkName, Channel binding)
	{
		partnerLinkBindings.put(partnerLinkName, binding);
		notifyAll();
	}

	/**
	 * Gets an incoming partner link's binding.  If the partner link is not yet
	 * bound, then this method blocks until it is bound.
	 * 
	 * @param partnerLink the partner link to get the binding for
	 * @return the partner link's binding
	 */
	synchronized Channel getIncomingLinkBinding(PartnerLink partnerLink)
	{
		while(!partnerLinkBindings.containsKey(partnerLink.getName()))
		{
			try
			{
				wait();
			}
			catch(InterruptedException e) { /**/ }
		}
		// Wait until the partner link has a binding
		// TODO: do we want to timeout?
		
		return (Channel)partnerLinkBindings.get(partnerLink.getName());
	}

	/**
	 * Gets an outgoing partner link's binding.
	 * 
	 * @param partnerLink the partner link to get the binding for
	 * @return the partner link's binding
	 */
	Object getOutgoingLinkBinding(PartnerLink partnerLink)
	{
		return server.getOutgoingBinding(partnerLink.getName());
	}
	
	/**
	 * Enables a partner link.
	 * 
	 * @param partnerLink the partner link to disable
	 */
	synchronized void enablePartnerLink(PartnerLink partnerLink)
	{
		enabledPartnerLinks.addElement(partnerLink.getName());
	}
	
	/**
	 * Disables a partner link.
	 * 
	 * @param partnerLink the partner link to disable
	 */
	synchronized void disablePartnerLink(PartnerLink partnerLink)
	{
		enabledPartnerLinks.removeElement(partnerLink.getName());
	}

	/**
	 * Gets whether or not a partner link is enabled.
	 * 
	 * @param partnerLinkName the name of the partner link to check
	 * @return whether or not the partner link is enabled
	 */	
	synchronized boolean isPartnerLinkEnabled(String partnerLinkName)
	{
		return enabledPartnerLinks.contains(partnerLinkName);
	}
	
	/**
	 * Attempts to handle a BPEL fault.
	 * 
	 * @param scopeData the process's current scope
	 * @param namespace the namespace of the fault's type
	 * @param name the fault's type
	 * @return whether or not the fault was caught by an in-scope handler
	 * 
	 * @throws BPELServerFault an error occured while executing the fault
	 * handler
	 */
	boolean handleFault(ScopeData scopeData, String namespace, String name)
		throws BPELServerFault
	{
		Catch catcch = scopeData.getCatch(namespace, name);
		if(catcch != null)
		{
			BPELServer.log.info("Executing <catch> handler for type "
				+ namespace + ":" + name);
			catcch.execute(this);
			return true;
		}
		// Try to get and execute a matching <catch> block
		
		CatchAll catchAll = scopeData.getCatchAll();
		if(catchAll != null)
		{
			BPELServer.log.info("Executing <catchAll> handler");
			catchAll.execute(this);
			return true;
		}
		// Otherwise, try the <catchAll> block
		
		BPELServer.log.severe("No handler found for type " + namespace + ":" + name);
		return false;
		// If all else fails, then we can't handle the fault
	}

	/**
	 * Gets part of a variable's value, in accordance with the
	 * <tt>bpws:getVariableData()</tt> XPath function.
	 * 
	 * @param scopeData the process's current scope
	 * @param variable the variable to look up
	 * @param partName the part of the variable to read (or <tt>null</tt> to
	 * read the entire variable) 
	 * @param locationPath the XPath location to read (or <tt>null</tt> for
	 * no location)
	 * @return an {@link Element} representation of the corresponding variable
	 * part
	 * 
	 * @throws BPELServerFault an error occured while trying to read the
	 * variable part
	 */
	public Element getVariablePart(ScopeData scopeData,
		VariableSpecification variable, String partName, String locationPath)
		throws BPELServerFault
	{
		Element source = getVariable(variable);
		// Get the variable
		if(source == null)
		{
			String error = "Variable " + variable.getName() + " not found";
			BPELServer.log.severe(error);
			if(!handleFault(scopeData, BPELServer.namespace,
				"uninitializedVariable"))
			{
				IllegalArgumentException e = new IllegalArgumentException(error);
				throw new BPELServerFault(e);
			}
			// If the variable has not value, throw an uninitializedVariable
			// fault

			return null;
		}
	
		// If the part name is specified
		if(partName != null)
		{
			source = source.getElement(null, partName);
			// Get the corresponding part
			if(source == null)
			{
				String error = "Variable part " + variable.getName() + "/"
					+ partName + " not found";
				BPELServer.log.severe(error);
				if(!handleFault(scopeData, BPELServer.namespace,
					"uninitializedVariable"))
				{
					IllegalArgumentException e = new IllegalArgumentException(
						error);
					throw new BPELServerFault(e);
				}

				return null;
			}
			// If the part is empty, then throw an uninitializedVariable
			// fault
		}
		
		// TODO: don't ignore locationPath
	
		return new Element(source);
		// Return a copy of the data
	}
}
