/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.ksoap2.*;
import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.soap.*;
import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;invoke&gt; tags.
 * 
 * @author Greg Hackmann
 */
class Invoke extends Transaction
{
	/**
	 * The variable that provides input to the invocation.
	 */
	private final VariableSpecification inputVariable;
	/**
	 * The variable that stores the invocation's output.
	 */
	private final VariableSpecification outputVariable;
	/**
	 * The invocation's compensation handler.
	 */
	private Activity compensationHandler = null;
	
	/**
	 * Creates a new Invoke.
	 * 
	 * @param parser the parser to read the invoke's attributes from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Invoke(XmlPullParser parser, ScopeData scopeData) throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		super(parser, new ScopeData(scopeData), "invoke");
		parseStartTag();
		
		inputVariable = getVariableSpecification("inputVariable");
		outputVariable = getVariableSpecification("outputVariable");
		// Read the input and output variables
		
		parser.nextTag();
		parseStandardElements();
		parseCorrelations(Correlation.UNSPECIFIED);
		// Parse the correlations (if any)
		
		Vector catchVector = new Vector();
		while("catch".equals(parser.getName()))
			catchVector.addElement(new Catch(parser, this.scopeData));
		// Parse all the child <catch>s

		Catch [] catches = new Catch[catchVector.size()];
		catchVector.copyInto(catches);
		this.scopeData.addCatches(catches);
		// Add them to the scope

		if("catchAll".equals(parser.getName()))
		{
			CatchAll catchAll = new CatchAll(parser, this.scopeData);
			this.scopeData.setCatchAll(catchAll);
		}
		// If there's a <catchAll>, parse it and add it to the scope

		if("compensationHandler".equals(parser.getName()))
		{
			parser.require(XmlPullParser.START_TAG, BPELServer.namespace,
				"compensationHandler");
			parser.nextTag();

			compensationHandler = Activity.parse(parser, scopeData);

			parser.require(XmlPullParser.END_TAG, BPELServer.namespace,
				"compensationHandler");
			parser.nextTag();
		}
		// If there's a <compensationHandler>, parse it
		
		parseEndTag();
		parser.nextTag();
	}

	boolean executeImpl(InstanceData instanceData) throws BPELServerFault
	{
		Object binding = partnerLink.getOutgoingBinding(instanceData);
		if(binding == null)
		{
			String error = "Partner link " + partnerLink +  " unbound";
			BPELServer.log.severe(error);
			IllegalArgumentException e = new IllegalArgumentException(error);
			throw new BPELServerFault(e);
		}
		Transport transport = instanceData.getTransport();
		// Get the binding and transport
		
		Element call = new Element();
		call.setName(operation);
		call.setNamespace(portTypeNamespace);
		// Create the call object
		
		if(inputVariable != null)
		{
			Element variable = instanceData.getVariable(inputVariable);
			int numChildren = variable.getChildCount();

			for(int i = 0; i < numChildren; i++)
				call.addChild(variable.getType(i), variable.getChild(i));
		}
		// If there's an input variable, copy its contents into the call
		
		Channel channel;
		try
		{
			channel = transport.openChannel(binding);
			BPELServer.log.info(toString() + " channel to " + binding +
				" opened");
		}
		catch(IOException e)
		{
			BPELServer.log.severe(toString() + ": IOException " + e +
				" thrown while opening channel to " + binding);
			e.printStackTrace();
			throw new BPELServerFault(e);
		}
		// Open a channel to the remote service
		
		synchronized(channel)
		{
			try
			{
				Element returnValue = channel.call(call);
				BPELServer.log.info(toString() + ": called remote service " +
					operation + "()");
	
				if(outputVariable != null)
					instanceData.setVariable(outputVariable, returnValue);
				
				// Call the remote service and copy the return value into the 
				// output variable, if any
			}
			// If the remote server threw a fault
			catch(SoapFault e)
			{
				BPELServer.log.warning(toString() + ": fault " + e +
					" caught during invoke");
				
				Element detail = (Element)e.detail.getChild(0);
				if(detail.getType(0) != Node.ELEMENT)
				{
					BPELServer.log.severe("Cannot understand fault");
					throw new BPELServerFault("Uncaught fault: " +
						e.faultstring, e);
				}
				// If the child isn't an element, then we can't handle the fault
				
				Element fault = (Element)detail.getChild(0);
				String faultName = fault.getName(), faultNamespace =
					fault.getNamespace();
				if(!instanceData.handleFault(scopeData, faultNamespace,
					faultName))
				{
					BPELServer.log.severe(toString() +
						": no handler found for type " + faultNamespace +
						":" + faultName);
					e.faultactor = binding.toString();
					throw new BPELServerFault("Uncaught fault: " +
						e.faultstring, e);
					// If no handler is found, pass along the fault
				}
			}
			catch(Throwable e)
			{
				BPELServer.log.severe(toString() + " exception " + e +
					" thrown during <invoke>");
				e.printStackTrace();
				throw new BPELServerFault(e);
				// Complain if any other exceptions are thrown
			}
			finally
			{
				try
				{
					if(!channel.isClosed())
					{
						channel.close();
						BPELServer.log.info(toString() + ": channel to " +
							binding + " closed");
					}
					// Close the channel at the end
				}
				catch(IOException e)
				{
					BPELServer.log.severe(toString() + ": IOException " + e +
						" thrown while closing channel to " + binding);
					e.printStackTrace();
					throw new BPELServerFault(e);
				}
			}
		}
		
		return true;
	}
}
