/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;pick&gt; tags.
 * 
 * @author Greg Hackmann
 */
class Pick extends Activity
{
	/**
	 * Whether or not this &lt;pick&gt; can start a new process.
	 */
	private final boolean createInstance;
	/**
	 * The &lt;onMessage&gt; child activities.
	 */
	final Vector onMessages = new Vector();
	/**
	 * The &lt;onAlarm&gt; child activities.
	 */
	final Vector onAlarms = new Vector();
	
	/**
	 * Creates a new Pick.
	 * 
	 * @param parser the parser to read the pick's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Pick(XmlPullParser parser, ScopeData scopeData) throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		super(parser, scopeData, "pick");
		parseStartTag();
		
		createInstance = "yes".equals(getAttribute("createInstance"));
		// Store the createInstance attribute
		
		parser.nextTag();
		parseStandardElements();
		
		while(parser.getName().equals("onMessage"))
			onMessages.addElement(new OnMessage(parser, scopeData));
		// Parse all the <onMessage> child tags
		
		if(onMessages.isEmpty())
			throw new MalformedBPELException(parser,
				"<pick> must specify at least one <onMessage>");
		// Complain if there aren't any <onMessage>s
		
		// TODO: parse onAlarms

		if(createInstance && onAlarms.size() > 1)
			throw new MalformedBPELException(parser,
				"<pick> with createInstance=\"yes\" cannot specify <onAlarm>s");
		// Complain if there's an incompatible combination of createInstance
		// and onAlarm children

		parseEndTag();
		parser.nextTag();
	}	

	Transaction [] getStartActivities()
	{
		if(createInstance)
		{
			Transaction [] startActivities = new Transaction[onMessages.size()];
			onMessages.copyInto(startActivities);
			return startActivities;
		}
		return new Receive[0];
		// Return the children if we're allowed to create an instance, otherwise
		// return nothing
	}

	boolean executeImpl(final InstanceData instanceData) throws BPELServerFault
	{
		MessageThread messageThread = new MessageThread(this, instanceData);
		
		messageThread.start();
		try
		{
			messageThread.join();
		}
		catch(InterruptedException e)
		{
			//
		}
		// Start a new thread to wait for a message to arrive, then wait for
		// the thread to die
		
		OnMessage fired = messageThread.getFiredEvent();
		if(fired != null)
		{
			BPELServer.log.info(toString() + ": child " + fired
				+ " fired; executing child activity");
			return fired.execute(instanceData);
		}
		// Get the <onMessage> child that fired (if any) and execute it
		
		return false;
		// TODO: handle alarms
	}
}

/**
 * Handles &lt;onMessage&gt; and &lt;onAlarm&gt; events for {@link Pick}.
 * We spawn off a seperate thread for this so that the alarms can interrupt
 * it; interrupting the main thread might disrupt something else.
 * 
 * @author Greg Hackmann
 */
class MessageThread extends Thread
{
	/**
	 * The {@link Pick} that started this thread.
	 */
	private final Pick parent;
	/**
	 * The current BPEL process instance.
	 */
	private final InstanceData instanceData;
	/**
	 * The {@link OnMessage} which received a message first.
	 */
	private OnMessage firedEvent;

	/**
	 * Creates a new MessageThread.
	 * 
	 * @param parent the {@link Pick} that started this thread
	 * @param instanceData the current BPEL process instance
	 */
	MessageThread(Pick parent, InstanceData instanceData)
	{
		super(parent + " MessageThread");
		this.parent = parent;
		this.instanceData = instanceData;
	}
	
	public void run()
	{
		OnMessage [] onMessageArray = new OnMessage[parent.onMessages.size()];
		parent.onMessages.copyInto(onMessageArray);
	
		synchronized(instanceData)
		{
			for(int i = 0; i < onMessageArray.length; i++)
				instanceData.enablePartnerLink(onMessageArray[i].partnerLink);
		}
		// Enable all the <onMessage>s' partner links
		try
		{
			BPELServer.log.info(toString() + ": waiting for message to arrive");
			firedEvent = instanceData.waitForMessage(onMessageArray);
		}
		catch(InterruptedException e)
		{
			BPELServer.log.info(toString()
				+ ": onMessage thread interrupted by onAlarm thread");
			firedEvent = null;
		}
		// Wait for a child to fire
		finally
		{
			synchronized(instanceData)
			{
				for(int i = 0; i < onMessageArray.length; i++)
					instanceData.disablePartnerLink(onMessageArray[i].partnerLink);
			}
		}
		// Disable all the <onMessage>s' partner links
	}
	
	/**
	 * Gets the {@link OnMessage} which received a message first.
	 * 
	 * @return the {@link OnMessage} which received a message first, or
	 * <tt>null</tt> if an alarm fired before any message arrived
	 */
	OnMessage getFiredEvent()
	{
		return firedEvent;
	}
}
