/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.soap.*;
import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents, and executes a BPEL process and manages all of its instances.
 * 
 * @author Greg Hackmann
 */
class Process
{
	/**
	 * The process's name.
	 */
	private final String name;
	/**
	 * The process's target namespace.
	 */
	private final String targetNamespace;
	/**
	 * Whether or not join failures should be ignored.
	 */
	private final boolean suppressJoinFailure;

	/**
	 * The &lt;partnerLinks&gt; child tag.
	 */
	private final PartnerLinks partnerLinks;
	/**
	 * The &lt;variables&gt; child tag.
	 */
	private final Variables variables;
	/**
	 * The &lt;correlationSets&gt; child tag.
	 */
	private final CorrelationSets correlationSets;
	/**
	 * The &lt;faultHandlers&gt; child tag.
	 */
	private final FaultHandlers faultHandlers;
	/**
	 * The process's "root" activity.
	 */
	private final Activity activity;
	/**
	 * All of the activities that can create a new process.
	 */
	private final Transaction [] startActivities;
	
	/**
	 * A list of process instances that are currently running.
	 */
	private final Vector activeInstances = new Vector();
	
	/**
	 * Parses a BPEL process.
	 * 
	 * @param parser the parser to read the BPEL process from
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Process(XmlPullParser parser)
		throws XmlPullParserException, IOException, MalformedDocumentException
	{
		ScopeData scopeData = new ScopeData();
		// Create an empty scope
		
		parser.require(XmlPullParser.START_TAG, BPELServer.namespace, "process");
		// Read the opening tag

		name = parser.getAttributeValue(null, "name");
		if(name == null)
			throw new MalformedBPELException(parser,
				"<process> must specify name");
		targetNamespace = parser.getAttributeValue(null, "targetNamespace");
		if(targetNamespace == null)
			throw new MalformedBPELException(parser,
				"<process> must specify targetNamespace");
		// Read the process's name and target namespace
		
		suppressJoinFailure = "yes".equals(parser.getAttributeValue(null,
			"suppressJoinFailure"));
		scopeData.setSuppressJoinFailure(suppressJoinFailure);
		// Read whether or not to ignore join failures
		
		// TODO: what about other attributes?
				
		parser.nextTag();

		partnerLinks = new PartnerLinks(parser);
		scopeData.setPartnerLinks(partnerLinks);
		// Read the <partnerLinks> child
//		partners = new Partners(parser);
		// TODO: parse partners
		variables = new Variables(parser);
		scopeData.addVariables(variables);
		// Read the <variables> child
		correlationSets = new CorrelationSets(parser);
		scopeData.addCorrelationSets(correlationSets);
		// Read the <correlationSets> child
		faultHandlers = new FaultHandlers(parser, scopeData);
		scopeData.addCatches(faultHandlers.getCatches());
		scopeData.setCatchAll(faultHandlers.getCatchAll());
		// Read the <faultHandlers> child
		// TODO: parse compensation handler
		// TODO: parse event handler
		activity = Activity.parse(parser, scopeData);
		startActivities = activity.getStartActivities();
		// Read the root activity
		
		parser.require(XmlPullParser.END_TAG, BPELServer.namespace, "process");
		// Read the closing tag
	}

	/**
	 * Gets the process's target namespace.
	 * 
	 * @return the process's target namespace
	 */
	String getTargetNamespace()
	{
		return targetNamespace;
	}
	
	/**
	 * Executes the BPEL process.
	 * 
	 * @param server the {@link BPELServer} that is hosting the process
	 * @param request a DOM representation of the incoming request message
	 * @param channel the {@link Channel} that the message arrived from
	 * 
	 * @throws BPELServerFault an error occured while the process was being
	 * executed
	 */
	void execute(BPELServer server, Element request, Channel channel)
		throws BPELServerFault
	{
		String namespace = request.getNamespace(), operation = request.getName();
		// Get the incoming request's name and namespace
		
		String partnerLinkName = server.getIncomingLink(namespace, operation);
		// Find out which partner link the request corresponds to
		if(partnerLinkName == null)
		{
			String error = "No partner link defined for operation " + operation;
			BPELServer.log.severe(error);
			IllegalArgumentException e = new IllegalArgumentException(error);
			throw new BPELServerFault(e);
			// Complain if there's no corresponding partner link
		}

		// For each start activities
		for(int i = 0; i < startActivities.length; i++)
		{
			// If the start activity's incoming partner link matches
			if(startActivities[i].partnerLink.getName().equals(partnerLinkName))
			{
				BPELServer.log.info("Incoming request matches partner link for "
					+ startActivities[i]);

				InstanceData newInstance = new InstanceData(server);
				newInstance.bindPartnerLink(startActivities[i].partnerLink,
					new ReplayChannel(channel, request));
				// Create a new process instance and bind the incoming request
				// to the appropriate partner link
				
				activeInstances.addElement(newInstance);
				BPELServer.log.info("Executing start activity " + activity);
				// Try to execute the process
				if(!activity.execute(newInstance))
				{
					activeInstances.removeElement(newInstance);
					String error = activity.toString() + " failed to execute";
					BPELServer.log.severe(error);
					throw new BPELServerFault(new Exception(error));
					// Complain if the start activity failed
				}
				
				activeInstances.removeElement(newInstance);
				return;
			}
		}
		
		Enumeration e = activeInstances.elements();
		// For each active instance
		while(e.hasMoreElements())
		{
			InstanceData instanceData = (InstanceData)e.nextElement();
			// If the corresponding partner link is enabled 
			if(instanceData.isPartnerLinkEnabled(partnerLinkName))
			{
				BPELServer.log.info("Incoming request matches enabled partner link for existing instance");
				instanceData.bindPartnerLink(partnerLinkName,
					new ReplayChannel(channel, request));
				return;
				// Bind the partner link to the incoming request
			}
		}
		
		String error = "No start activity or active instance found for partner link " +
			partnerLinkName;
		BPELServer.log.severe(error);
		IllegalArgumentException ex = new IllegalArgumentException(error);
		throw new BPELServerFault(ex);
		// Complain if no matching start activity was found
	}
}

/**
 * A special {@link Channel} which "replays" the incoming request message.
 * Using this, the BPEL server can receive the full request and send it to
 * the proper process, without the corresponding {@link Receive} activity
 * being aware that the message has already been consumed.
 * 
 * @author Greg Hackmann
 */
class ReplayChannel extends Channel
{
	/**
	 * The incoming request.
	 */
	private final Element request;
	/**
	 * The {@link Channel} that the request arrived from.
	 */
	private final Channel channel;

	/**
	 * Creates a new ReplayChannel.
	 * 
	 * @param channel the {@link Channel} that the request arrived from
	 * @param request the incoming request
	 */
	ReplayChannel(Channel channel, Element request)
	{
		this.channel = channel;
		this.request = request;
	}

	public void closeImpl() throws IOException
	{
		channel.close();
	}

	public Object receiveObject() throws IOException, XmlPullParserException,
		MustUnderstandException
	{
		return null;
		// Note that this will never be called, since <receive> and <invoke>
		// only call receiveElement()
	}

	public Element receiveElement() throws IOException, XmlPullParserException,
		MustUnderstandException
	{
		return request;
	}

	public void sendObject(Object object) throws IOException
	{
		channel.sendObject(object);
	}

	public void sendElement(Element element) throws IOException
	{
		channel.sendElement(element);
	}
}
