/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.soap.*;
import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;receive&gt; tags.
 * 
 * @author Greg Hackmann
 */
class Receive extends Transaction
{
	/**
	 * The variable that the message will be stored in.
	 */
	private final VariableSpecification variableSpec;
	/**
	 * Whether or not this &lt;receive&gt; can start a new process.
	 */
	protected boolean createInstance = false;
	
	/**
	 * Creates a new Receive.
	 * 
	 * @param parser the parser to read the receive's attributes from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Receive(XmlPullParser parser, ScopeData scopeData) throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		this(parser, scopeData, "receive");
	}

	/**
	 * Creates a new Receive with a specified tag name.
	 * 
	 * @param parser the parser to read the receive's attributes from
	 * @param scopeData a description of the current scope
	 * @param tagName the name of the tag (usually "receive")
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 * 
	 * @see OnMessage#OnMessage(XmlPullParser, ScopeData)
	 */
	protected Receive(XmlPullParser parser, ScopeData scopeData,
		String tagName) throws IOException, XmlPullParserException,
		MalformedDocumentException
	{
		super(parser, scopeData, tagName);
		parseStartTag();

		variableSpec = getVariableSpecification("variable");
		createInstance = "yes".equals(getAttribute("createInstance"));
		// Store the variableSpec and createInstance attributes

		parser.nextTag();
		parseBody();
		
		parseEndTag();
		parser.nextTag();
	}
	
	/**
	 * Parses the Receive's body.
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	protected void parseBody() throws IOException, XmlPullParserException,
		MalformedDocumentException
	{
		parseStandardElements();
		parseCorrelations(Correlation.IN);	
		// Parse the standard parts, then the correlations
	}

	Transaction [] getStartActivities()
	{
		if(createInstance)
			return new Transaction [] { this };
		return new Transaction[0];
		// Return this if we're allowed to create an instance, otherwise
		// return nothing
	}
	
	boolean executeImpl(InstanceData instanceData) throws BPELServerFault
	{
		try
		{
			Channel channel;
			
			synchronized(instanceData)
			{
				instanceData.enablePartnerLink(partnerLink);
				channel = instanceData.getIncomingLinkBinding(partnerLink);
				instanceData.disablePartnerLink(partnerLink);
			}
			// Enable the partner link, get the binding, and then disable it
			// again
			
			if(channel == null)
			{
				String error = "Partner link " + partnerLink +  " unbound";
				BPELServer.log.severe(error);
				IllegalArgumentException e =
					new IllegalArgumentException(error);
				throw new BPELServerFault(e);
			}
			// If the partner link is null, then it's unbound, so throw an error
			
			Element returnValue = channel.receiveElement();
			BPELServer.log.info(toString() + ": received remote value");
			if(variableSpec != null)
				instanceData.setVariable(variableSpec, returnValue);
			// Receive the message from the partner link and update the
			// variable (if applicable)
			
			return true;			
		}
		catch(Exception e)
		{
			BPELServer.log.severe(toString() + ": exception " + e +
				" thrown while receiving value");
			e.printStackTrace();
			throw new BPELServerFault(e);
		}
	}
}
