/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

/**
 * Represents a BPEL process scope.
 * 
 * @author Greg Hackmann
 */
public class ScopeData
{
	/**
	 * The {@link Links} in the current scope.
	 */
	private Links [] linkScope;
	/**
	 * The {@link Variables} in the current scope.
	 */
	private Variables [] variableScope;
	/**
	 * The {@link CorrelationSets} in the current scope.
	 */
	private CorrelationSets [] correlationSetScope;
	/**
	 * The {@link Catch}es in the current scope.
	 */
	private Catch [] catchScope;
	/**
	 * The current scope's {@link CatchAll}.
	 */
	private CatchAll catchAllScope;
//	private final CompensationHandlers [] compensationHandlerScope;
//	private final EventHandlers [] eventHandlerScope;
	/**
	 * The {@link PartnerLinks} in the current scope.
	 */
	private PartnerLinks partnerLinks;
	/**
	 * Whether or not join failures should be ignored by default.
	 */
	private boolean suppressJoinFailure = false;

	/**
	 * Creates a new empty ScopeData.
	 */
	ScopeData()
	{
		linkScope = new Links[0];
		variableScope = new Variables[0];
		correlationSetScope = new CorrelationSets[0];
		catchScope = new Catch[0];
		catchAllScope = null;
	}
	
	/**
	 * Creates a new ScopeData.
	 * 
	 * @param linkScope the {@link Links} in the new scope
	 * @param variableScope the {@link Variables} in the new scope
	 * @param correlationSetScope the {@link CorrelationSets} in the new scope
	 * @param catchScope the {@link Catch}es in the new scope
	 * @param catchAllScope the new scope's {@link CatchAll}
	 * @param partnerLinks the {@link PartnerLinks} in the new scope
	 * @param suppressJoinFailure whether or not activities in the new scope
	 * should ignore link join failures
	 */
	protected ScopeData(Links [] linkScope, Variables [] variableScope,
		CorrelationSets [] correlationSetScope, Catch [] catchScope,
		CatchAll catchAllScope, PartnerLinks partnerLinks,
		boolean suppressJoinFailure)
	{
		this.linkScope = linkScope;
		this.variableScope = variableScope;
		this.correlationSetScope = correlationSetScope;
		this.catchScope = catchScope;
		this.catchAllScope = catchAllScope;
		this.partnerLinks = partnerLinks;
		this.suppressJoinFailure = suppressJoinFailure;
	}
	
	/**
	 * Copies an existing ScopeData.
	 * 
	 * @param other the ScopeData to copy
	 */
	ScopeData(ScopeData other)
	{
		this.linkScope = other.linkScope;
		this.variableScope = other.variableScope;
		this.correlationSetScope = other.correlationSetScope;
		this.catchScope = other.catchScope;
		this.catchAllScope = other.catchAllScope;
		this.partnerLinks = other.partnerLinks;
		this.suppressJoinFailure = other.suppressJoinFailure;
	}

	/**
	 * Copies this ScopeData, and removes the inter-activity links from the
	 * copy. This is used by {@link While}, since links are not allowed to
	 * cross the scope defined by a &lt;while&gt; tag.
	 * 
	 * @return a copy of this ScopeData, without any links
	 */
	ScopeData removeLinks()
	{
		return new ScopeData(new Links[0], variableScope, correlationSetScope,
			catchScope, catchAllScope, partnerLinks, suppressJoinFailure);
	}
	
	/**
	 * Adds inter-activity links to the current scope.
	 * 
	 * @param newLinks the new {@link Links} to add to the scope
	 */
	void addLinks(Links newLinks)
	{
		Links [] newLinkScope = new Links[linkScope.length + 1];
		concatenate(newLinks, linkScope, newLinkScope);
		linkScope = newLinkScope;
		// Concatenate the old and new links, then update the scope
	}

	/**
	 * Gets an inter-activity link.
	 * 
	 * @param name the link's name
	 * @return the {@link LinkSpecification} that represents the corresponding
	 * link, or <tt>null</tt> if there is no such link in this scope
	 */
	LinkSpecification getLink(String name)
	{
		// For each link collection in the scope
		for(int i = 0; i < linkScope.length; i++)
		{
			Links links = linkScope[i];
			if(links.containsKey(name))
				return links.getLinkSpecification(name);
			// If the collection has the link, then return it
		}
		
		return null;
	}
	
	/**
	 * Adds variables to the current scope.
	 * 
	 * @param newVariables the new {@link Variables} to add to the scope
	 */
	void addVariables(Variables newVariables)
	{
		Variables [] newVariableScope = new Variables[variableScope.length + 1];
		concatenate(newVariables, variableScope, newVariableScope);
		variableScope = newVariableScope;
		// Concatenate the old and new variables, then update the scope
	}
	
	/**
	 * Gets a variable.
	 * 
	 * @param name the variables's name
	 * @return the {@link VariableSpecification} that represents the
	 * corresponding variable, or <tt>null</tt> if there is no such variable
	 * in this scope
	 */
	public VariableSpecification getVariable(String name)
	{
		// For each variable collection in the scope
		for(int i = 0; i < variableScope.length; i++)
		{
			Variables variables = variableScope[i];
			if(variables.containsKey(name))
				return variables.getVariableSpecification(name);
			// If the collection has the variable, then return it
		}
		
		return null;
	}

	/**
	 * Adds correlation sets to the current scope.
	 * 
	 * @param newCorrelationSets the new {@link CorrelationSets} to add to the scope
	 */
	void addCorrelationSets(CorrelationSets newCorrelationSets)
	{
		CorrelationSets [] newCorrelationSetScope = new
			CorrelationSets[correlationSetScope.length + 1];
		concatenate(newCorrelationSets, correlationSetScope,
					newCorrelationSetScope);
		correlationSetScope = newCorrelationSetScope;
		// Concatenate the old and new correlation sets, then update the scope
	}
	
	/**
	 * Adds fault handlers to the current scope.
	 * 
	 * @param newCatches the new {@link Catch}es to add to the scope
	 */
	void addCatches(Catch [] newCatches)
	{
		Catch [] newCatchScope = new Catch[catchScope.length
			+ newCatches.length];
		System.arraycopy(newCatches, 0, newCatchScope, 0, newCatches.length);
		System.arraycopy(catchScope, 0, newCatchScope, newCatches.length,
							catchScope.length);

		catchScope = newCatchScope;
		// Concatenate the old and new handlers, then update the scope
	}
	
	/**
	 * Gets the fault handler for the specified fault type.
	 * 
	 * @param namespace the namespace of the fault type
	 * @param name the fault type
	 * @return the corresponding fault handler, or <tt>null</tt> if there is
	 * no fault handler for the specified fault type
	 */
	Catch getCatch(String namespace, String name)
	{
		// For each handler in the scope
		for(int i = 0; i < catchScope.length; i++)
		{
			Catch catcch = catchScope[i];
			if(catcch.matchesType(namespace, name))
				return catcch;
			// Return it if it matches
		}
		
		return null;
	}

	/**
	 * Sets the scope's default fault handler.
	 * 
	 * @param catchAll the new default fault handler
	 */
	void setCatchAll(CatchAll catchAll)
	{
		catchAllScope = catchAll;
	}
	
	/**
	 * Gets the scope's default fault handler.
	 * 
	 * @return the scope's default handler, or <tt>null</tt> if none is
	 * defined
	 */
	CatchAll getCatchAll()
	{
		return catchAllScope;
	}
	
	/**
	 * Sets the scope's partner links.
	 * 
	 * @param partnerLinks the scope's new {@link PartnerLinks}
	 */
	void setPartnerLinks(PartnerLinks partnerLinks)
	{
		this.partnerLinks = partnerLinks;
	}
	
	/**
	 * Gets a partner link.
	 * 
	 * @param name the partner links's name
	 * @return the {@link PartnerLink} that represents the corresponding partner
	 * link, or <tt>null</tt> if there is no such partner link in this scope
	 */
	PartnerLink getPartnerLink(String name)
	{
		return partnerLinks.getPartnerLink(name);
	}

	/**
	 * Sets whether or not join failures should be ignored by default.
	 * 
	 * @param suppressJoinFailure whether or not join failures should be ignored
	 * by default
	 */
	void setSuppressJoinFailure(boolean suppressJoinFailure)
	{
		this.suppressJoinFailure = suppressJoinFailure;
	}

	/**
	 * Gets whether or not join failures should be ignored by default.
	 * 
	 * @return whether or not join failures should be ignored by default
	 */
	boolean isSuppressJoinFailure()
	{
		return suppressJoinFailure;
	}
	
	/**
	 * Copies an array and adds a new element to the beginning of the copy. It
	 * is assumed that <tt>newArray</tt> and <tt>oldArray</tt> are arrays of
	 * type <tt>addMe.class</tt>, and that
	 * <tt>newArray.length == oldArray.length + 1</tt>.
	 * 
	 * @param addMe the new element to add to the array
	 * @param oldArray the array to copy from
	 * @param newArray the array to copy into
	 */
	private void concatenate(Object addMe, Object [] oldArray, Object [] newArray)
	{
		newArray[0] = addMe;
		System.arraycopy(oldArray, 0, newArray, 1, oldArray.length);
	}
}
