/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;
import java.util.*;

import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;sequence&gt; tags.
 * 
 * @author Greg Hackmann
 */
class Sequence extends Activity
{
	/**
	 * A list of all the child activities.
	 */
	private final Vector children = new Vector();
	
	/**
	 * Creates a new Sequence.
	 * 
	 * @param parser the parser to read the sequence's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */
	Sequence(XmlPullParser parser, ScopeData scopeData) throws IOException,
		XmlPullParserException, MalformedDocumentException
	{
		super(parser, scopeData, "sequence");
		parseStartTag();
		
		parser.nextTag();
		parseStandardElements();
		
		while(parser.getEventType() != XmlPullParser.END_TAG)
			children.addElement(Activity.parse(parser, scopeData));
		// Keep adding children until we reach the </sequence> tag
		
		if(children.isEmpty())
			throw new MalformedBPELException(parser,
				"<sequence> must specify at least one activity");
		// Complain if there's not at least one child
		
		parseEndTag();
		parser.nextTag();
	}

	Transaction [] getStartActivities()
	{
		Activity first = (Activity)children.firstElement();
		return first.getStartActivities();
		// Only the first activity in the sequence can have start activities
	}
	
	boolean executeImpl(InstanceData instanceData) throws BPELServerFault
	{
		Enumeration e = children.elements();

		// For all child activities
		while(e.hasMoreElements())
		{
			Activity next = (Activity)e.nextElement();
			boolean succeeded = true;
			// Try to execute the next task
			try
			{
				BPELServer.log.info(toString() + ": executing child " + next);
				succeeded = next.execute(instanceData);
			}
			// If the task fails
			catch(BPELServerFault f)
			{
				BPELServer.log.severe(toString() + ": child " + next +
					" threw a fault");
				// For all remaining child activities
				while(e.hasMoreElements())
				{
					try
					{
						Activity a = (Activity)e.nextElement();
						a.fireLinks(instanceData, false);
					}
					catch(BPELServerFault dummy) { /**/ }
					// Fire the links as false.  If something goes wrong, ignore
					// it, since we have a more important fault to throw.
				}
				
				throw f;
				// Rethrow the fault
			}
			
			// If we need to do dead path elimination
			if(!succeeded)
			{
				BPELServer.log.severe(toString() + ": child " + next +
					" failed");

				BPELServerFault linkFault = null;
				
				//	For all remaining child activities
				while(e.hasMoreElements())
				{
					try
					{
						Activity a = (Activity)e.nextElement();
						a.fireLinks(instanceData, false);
						// Fire the link as false
					}
					catch(BPELServerFault f)
					{
						if(linkFault != null)
							linkFault = f;
					}
					// If any links throw a fault, then catch the first one.
					// (Note that we want to keep going even if one of the links
					// blows up, since we need to notify the whole path that
					// it's being eliminated.)
				}
				
				if(linkFault != null)
					throw linkFault;
				// Re-throw any caught link faults
				
				return false;
				// Finally, pass along the failure
			}
		}
		
		return true;
	}
}
