/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Represents and executes BPEL &lt;to&gt; tags.
 * 
 * @author Greg Hackmann
 */
class To
{
	/**
	 * The destination of the data that will be written.
	 */
	private final ToSink sink;
	
	/**
	 * Creates a new To.
	 * 
	 * @param parser the parser to read the to's attributes and children
	 * from
	 * @param scopeData a description of the current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedDocumentException the BPEL parser read malformed data
	 */	
	To(XmlPullParser parser, ScopeData scopeData)
		throws XmlPullParserException, IOException, MalformedDocumentException
	{
		parser.require(XmlPullParser.START_TAG, BPELServer.namespace, "to");
		
		if(hasAttribute(parser, "variable"))
			sink = new VariablePartSink(parser, scopeData);
		// First try variable sinks
		else
			throw new MalformedBPELException(parser, "<to> has invalid to-spec");
		
		// TODO:
		// <to partnerLink="ncname"/>
		// <to variable="ncname" property="qname"/>

		parser.require(XmlPullParser.END_TAG, BPELServer.namespace, "to");
		parser.nextTag();
	}

	/**
	 * Reads the data to the specified destination.
	 * 
	 * @param value an {@link Element} representation of the data to write
	 * @param instanceData the current BPEL process instance
	 * 
	 * @throws BPELServerFault an error occured while writing the data
	 */
	void write(Element value, InstanceData instanceData)
		throws BPELServerFault
	{
		sink.write(value, instanceData);
	}

	/**
	 * Determines whether or not the tag has the specified attribute.
	 * 
	 * @param parser the parser that is reading the current tag
	 * @param name the name of the attribute to look for
	 * @return whether or not the tag has the specified attribute
	 */
	private boolean hasAttribute(XmlPullParser parser, String name)
	{
		return parser.getAttributeValue(null, name) != null;
	}
}
