/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.bpel;

import java.io.*;

import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * Describes the destination for a {@link To} operation.
 * 
 * @author Greg Hackmann
 */
interface ToSink
{
	/**
	 * Reads data to the current sink.
	 * 
	 * @param value an {@link Element} representation of the data to write
	 * @param instanceData the current BPEL process instance
	 * 
	 * @throws BPELServerFault an error occurred while writing the data
	 */
	void write(Element value, InstanceData instanceData)
		throws BPELServerFault;
}

/**
 * Describes data sinks specified by BPEL variable parts.
 * 
 * @author Greg Hackmann
 */
class VariablePartSink implements ToSink
{
	/**
	 * The BPEL variable that the data will be written to.
	 */
	private final VariableSpecification variable;
	/**
	 * The part of the variable that will be written.
	 */
	private final String partName;
	
	/**
	 * Creates a new VariablePartSink.
	 * 
	 * @param parser the parser that the expression comes from
	 * @param scopeData the process's current scope
	 * 
	 * @throws IOException there was an I/O error while reading from the parser
	 * @throws XmlPullParserException the XML parser read malformed data
	 * @throws MalformedBPELException the BPEL parser read malformed data
	 */
	VariablePartSink(XmlPullParser parser, ScopeData scopeData)
		throws MalformedBPELException, IOException, XmlPullParserException
	{
		String part = parser.getAttributeValue(null, "part");
		if(part == null)
			partName = null;
		else
		{
			String [] splitPart = Namespace.expandNamespace(part, parser);
			partName = splitPart[1];
		}
		// Read the part, and split off the name portion if the part isn't empty
		
		String variableName = parser.getAttributeValue(null, "variable");
		if(variableName == null)
			throw new MalformedBPELException(parser,
				"<to> must specify variable");
		variable = scopeData.getVariable(variableName);
		// Read the variable name and get the corresponding spec		

		parser.nextTag();
	}
	
	public void write(Element value, InstanceData instanceData)
		throws BPELServerFault
	{
		// If no part was specified
		if(partName == null)
		{
			instanceData.setVariable(variable.getName(), value);
			BPELServer.log.info(toString() + ": setting variable " +
				variable.getName());
			// Write directly into the variable
		}
		// Otherwise, if a part was specified
		else
		{
			Element oldValue = instanceData.getVariable(variable.getName());
			if(oldValue == null)
			{
				oldValue = new Element();
				oldValue.setName(variable.getMessageType());
				oldValue.setNamespace(variable.getMessageTypeNamespace());
				BPELServer.log.info(toString() + ": creating variable " +
					variable.getName());
			}
			// Get the old value if it exists, or make a new one if it doesn't

			value.setName(partName);
			int oldChild = oldValue.indexOf(null, partName, 0);
			if(oldChild >= 0)
				oldValue.removeChild(oldChild);
			// Remove the old part, if it exists
			
			oldValue.addChild(Node.ELEMENT, value);
			instanceData.setVariable(variable.getName(), oldValue);
			BPELServer.log.info(toString() + ": updating variable " +
				variable.getName());
			// Add the new part and update the variable's value
		}
	}
}
