/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.soap.midp;

import org.ksoap2.serialization.*;

import edu.wustl.mobilab.sliver.soap.*;

/**
 * Delegates SOAP requests to Java object.
 * To use this handler, create a class which will handle all of the
 * requests have a particular namespace.  This class must implement
 * {@link SOAPServiceProvider}.  Then, create methods
 * in this class named after each request, using the same
 * parameters and return values specified by the WSDL file.
 * Finally, implement the {@link SOAPServiceProvider#invoke(String, Object[]) invoke()}
 * method, which delegates the requests the class receives to the appropriate
 * methods.
 * 
 * <p>Because of the way SOAP handles variable names, all return values
 * must implement the {@link KvmSerializable} interface.  I suggest extending
 * the {@link SoapObject} helper class to do this; see the code snippet below
 * for more information.</p>
 * 
 * <p>If one of your requests uses a non-standard message type
 * (i.e., something that's not defined by XSD or SOAP), then be sure
 * to use the {@link Serialization#registerClass(String, Class)} to
 * register this custom type with Sliver.</p>
 * 
 * <p>Example usage:</p>
 * 
 * <pre>
 *public class AddResponse extends SoapObject
 *{
 *    public AddResponse()
 *    {
 *        this(0.0);
 *    }
 *
 *    public AddResponse(double result)
 *    {
 *        super("http://example.com/SampleService", "AddResponse");
 *        addProperty("result", new Double(result));
 *    }
 *}
 *
 *public class SampleService implements SOAPServiceProvider
 *{
 *    public AddResponse add(double in1, double in2)
 *    {
 *        return new AddResponse(in1 + in2);
 *    }
 *    
 *    public Object invoke(String methodName, Object [] params) throws Exception
 *    {
 *        if(methodName.equals("add"))
 *        {
 *            Double in1 = (Double)params[0], in2 = (Double)params[1];
 *            add(in1.doubleValue(), in2.doubleValue());
 *        }
 *        else
 *            throw new Exception("No method named " + methodName + " found");
 *    }
 *}
 *
 *...
 *
 *Serialization.registerClass("http://example.com/SampleService", AddResponse.class);
 *server.registerService("http://example.com/SampleService", new MethodCallHandler(new SampleService()));</pre>
 */
public class MethodCallHandler implements SOAPInvocationHandler
{
	/**
	 * The {@link SOAPServiceProvider} that will handle incoming requests.
	 */
	private final SOAPServiceProvider provider;
	
	/**
	 * Creates a new MethodCallHandler.
	 * 
	 * @param provider the {@link SOAPServiceProvider} that will handle
	 * incoming requests
	 */
	public MethodCallHandler(SOAPServiceProvider provider)
	{
		this.provider = provider;
	}	

	public Object handleInvocation(SoapObject request) throws
		IllegalArgumentException, SerializableException, UnserializableException
	{
		Object [] params = new Object[request.getPropertyCount()];
		for(int i = 0; i < params.length; i++)
			params[i] = request.getProperty(i);
		// Copy the parameters to a list
	
		try
		{
			return provider.invoke(request.getName(), params);
		}
		catch(SerializableException e)
		{
			throw e;
		}
		catch(Throwable t)
		{
			throw new
				UnserializableException("Exception thrown while invoking " +
					request.getName(), t);
		}
		// Invoke the provider, re-throwing any exceptions in the proper form
	}
}
